# BSD 3-Clause License
#
# Copyright (c) 2021, Austin Cummings
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# todo alt_max should be user input

import numpy as np

# Constants to describe the atmospheric density profile as described by Linsley

# Maximum altitude of the atmosphere. Above this, density is 0g/cm^3
z_max = 112.79

# Layers of the atmosphere where the behavior changes (in km)
atmos_layer_boundaries = np.array([4, 10, 40, 100, z_max])

# Model parameters (b in g/cm^2, c in cm)
atmos_b = np.array([1222.6562, 1144.9069, 1305.5948, 540.1778, 1])
atmos_c = np.array([994186.38, 878153.55, 636143.04, 772170.16, 1e9])


# Avagadro's Constant (molecules/mol)
N_A = 6.022e23

# Average Molar Mass of Air (kg/mol)
mm_air = 28.9647e-3

# Average radius of the Earth (km)
R_Earth = 6378  # km

# Maximum altitude of the atmosphere to consider (km)
alt_max = 112.79

# Electron rest mass (MeV)
electron_mass = 0.511

# Electron charge (in Coulombs)
electron_charge = 1.602e-19

# Conversion factor between rest mass (in MeV/c^2) and kg
MeV_to_kg = 1.79E-30

# Fine structure constant
alpha_fsc = 1 / 137

# Speed of light (km/s)
cl = 3e5



# BSD 3-Clause License
#
# Copyright (c) 2021, Austin Cummings
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# todo: Add folder output and save figures

from enum import IntEnum
import numpy as np
import matplotlib.pyplot as plt
# from matplotlib  import cm
# from labellines import labelLine, labelLines

# from scipy import interpolate
# from scipy.interpolate import griddata
# from scipy.optimize import curve_fit

# from . import atmosphere as atm
# from . import geometry as geo
# from . import constants as const
# from . import fit_func as fitf

plt.rcParams['axes.labelsize'] = 13
plt.rcParams['xtick.labelsize'] = 11
plt.rcParams['ytick.labelsize'] = 11
plt.rcParams['legend.fontsize'] = 11
plt.rcParams.update({'figure.autolayout': True})


class Plot(IntEnum):
    """
    Enum to identify which plots to produce
    """
    none = 0
    basic = 2
    general = 3
    all = 4


###################################################################
# General Plots

def plot_wavelength_distribution(wavelengths, photons):
    """Plots the wavelength distribution.
    Plots the differential wavelength distribution (non-normalized) of the Cherenkov photons arriving on the detection plane
    Because the angles involved are small, the distribution should be largely the same viewing anywhere on the detection plane

    Parameters
    ----------
    wavelengths: [array]
        [description]
    photons : [type]
        [description]

    Returns
    -------
    ArrayLike
        ptical depth (tau: unitless)
    """

    fig = plt.figure()
    plt.plot(wavelengths, photons)
    plt.xlabel('Wavelength (nm)')
    plt.ylabel(r'$\frac{dN_{\gamma}}{d\lambda}$')
    plt.grid(which='both')

    return fig


def plot_3D_spatial_distribution(r_bins, det_array):
    # Plots the 3-Dimensional spatial distribution (per m^2) of the Cherenkov photons arriving on the detection plane

    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    ax1.set_xlabel('X (km)')
    ax1.set_ylabel('Y (km)')
    im = ax1.imshow(det_array, interpolation='nearest', origin='lower', extent=[r_bins[0], r_bins[-1], r_bins[0], r_bins[-1]])
    fig.colorbar(im, ax=ax1)

    return fig


def plot_photon_time_spread(r_bins, time_bins, time_counts):
    # Plots the time spread within which 90% of the arriving photons arrive (in s), per spatial bin

    # Calculate the upper and lower time bounds which contain the data
    cumulative_counts = np.cumsum(time_counts, axis=1) / np.sum(time_counts, axis=1)[:, None]
    low_bin, high_bin = np.argmax(cumulative_counts >= 0.05, axis=1), np.argmax(cumulative_counts >= 0.95, axis=1)
    time_diffs = time_bins[high_bin] - time_bins[low_bin]

    fig = plt.figure()
    plt.plot(r_bins[:-1], time_diffs)
    plt.yscale('log')
    plt.xlabel('Distance From Shower Axis (km)')
    plt.ylabel('90% Time Spread (ns)')
    plt.grid(which='both')

    return fig


def plot_photon_angular_spread(r_bins, angular_bins, angular_counts):
    # Plots the angular spread within which 90% of the arriving photons arrive (in degrees), per spatial bin

    # Calculate the upper and lower angular bounds which contain the data
    cumulative_counts = np.cumsum(angular_counts, axis=1) / np.sum(angular_counts, axis=1)[:, None]
    low_bin, high_bin = np.argmax(cumulative_counts >= 0.05, axis=1), np.argmax(cumulative_counts >= 0.95, axis=1)
    angular_diffs = angular_bins[high_bin] - angular_bins[low_bin]

    fig = plt.figure()
    plt.plot(r_bins[:-1], angular_diffs)
    plt.yscale('log')
    plt.xlabel('Distance From Shower Axis (km)')
    plt.ylabel('90% Angular Spread (degrees)')
    plt.grid(which='both')

    return fig



# BSD 3-Clause License
#
# Copyright (c) 2021, Austin Cummings
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' Geometric Functions

.. autosummary::
   :toctree:
   :recursive:

   Earth_limb_theta_D
   theta_D_from_theta_EE
   theta_EE_from_theta_D
   theta_AE_from_theta_D
   z_from_L
   L_from_z
   ellipse_focii
   ellipse_coordinates

'''

import numpy as np
from . import constants as const


def Earth_limb_theta_D(det_alt):
    """Calculates the detector viewing angle (the angle between the trajectory and the detector doward zenith: in degrees)
       where the Earth's limb occurs given the detector altitude (in km)

    Parameters
    ----------
    det_alt : float
        detector altitude in km

    Returns
    -------
    float
        detector viewing angle in degree
    """
    theta_D = (180 / np.pi) * np.arcsin(const.R_Earth / (const.R_Earth + det_alt))

    return theta_D


def theta_D_from_theta_EE(det_alt, theta_EE):
    """Calculates the detector viewing angle (the angle between the trajectory and the detector doward zenith: in degrees)
    given the Earth emergence angle (the angle between the Earth horizontal and the trajectory: in degrees)
    and the altitude of the detector (in km)

    Parameters
    ----------
    det_alt : float
        detector altitude in km
    theta_EE : [type]
        earth emergence angle in degree

    Returns
    -------
    float
        detector viewing angle in degree
    """
    theta_EE = theta_EE * (np.pi / 180)
    theta_D = (180 / np.pi) * np.arcsin((const.R_Earth / (const.R_Earth + det_alt)) * np.sin(np.pi / 2 + theta_EE))

    return theta_D


def theta_EE_from_theta_D(det_alt, theta_D):
    """Calculates the Earth emergence angle (the angle between the Earth horizontal and the trajectory: in degrees)
    of the below limb event given the detector viewing angle (the angle between the trajectory and the detector doward zenith: in degrees)
    and altitude of the detector (in km)

    Parameters
    ----------
    det_alt : float
        detector altitude in km
    theta_D : [type]
        detector viewing angle in degree

    Returns
    -------
    float
        Earth emergence angle in degree
    """
    theta_D = theta_D * (np.pi / 180)
    theta_EE = (180 / np.pi) * (np.arccos(((const.R_Earth + det_alt) / const.R_Earth) * np.sin(theta_D)))

    return theta_EE


def theta_AE_from_theta_D(det_alt, theta_D):
    """Calculates the atmosphere entrance angle (the angle between the downwards CR trajectory and the atmosphere normal: in degrees)
    given the detector altitude (in km) and the detector viewing angle (the angle between the trajectory and the detector doward zenith: in degrees)

    Parameters
    ----------
    det_alt : float
        detector altitude in km
    theta_D : float
        detector viewing angle in degree

    Returns
    -------
    float
        atmospheric enrtance angle in degree
    """
    theta_D = theta_D * (np.pi / 180)
    theta_AE = (180 / np.pi) * np.arcsin(((const.R_Earth + det_alt) / (const.R_Earth + const.alt_max)) * np.sin(theta_D))

    return theta_AE


def z_from_L(L, theta_D, det_alt):
    """Calculates the altitude (in km) given the distance along the trajectory (in km) and the
    detector viewing angle (the angle between the trajectory and the detector doward zenith: in degrees)

    Parameters
    ----------
    L : float
        distance alog shower axis in km
    theta_D : float
        detetcor viewing angle in degree
    det_alt : float
        detector altitude in km

    Returns
    -------
    float
        altitude in km
    """
    theta_D_Earth_limb = Earth_limb_theta_D(det_alt)

    if theta_D > theta_D_Earth_limb:
        theta_AE = theta_AE_from_theta_D(det_alt, theta_D) * (np.pi / 180)
        z = np.sqrt((const.R_Earth + const.alt_max)**2 + L**2 - 2 * L * (const.R_Earth + const.alt_max) * np.cos(theta_AE)) - const.R_Earth
    else:
        theta_EE = theta_EE_from_theta_D(det_alt, theta_D) * (np.pi / 180)
        z = np.sqrt(const.R_Earth**2 + L**2 + 2 * const.R_Earth * L * np.sin(theta_EE)) - const.R_Earth

    return z


def L_from_z(z, theta_D, det_alt):
    """Calculates the distance along the trajectory (in km) given an altitude (in km) and the
    detector viewing angle (the angle between the trajectory and the detector doward zenith: in degrees)

    Parameters
    ----------
    z : float
        altitude in km
    theta_D : float
        detetcor viewing angle
    det_alt : float
        detetcor altitude

    Returns
    -------
    float
        distance along shower axis in km
    """
    theta_D_Earth_limb = Earth_limb_theta_D(det_alt)

    if theta_D > theta_D_Earth_limb:
        theta_AE = theta_AE_from_theta_D(det_alt, theta_D) * (np.pi / 180)
        a = 1
        b = -2 * (const.R_Earth + const.alt_max) * np.cos(theta_AE)
        c = (const.R_Earth + const.alt_max)**2 - (const.R_Earth + z)**2
        root_1 = (-b - np.sqrt(b**2 - 4 * a * c)) / (2 * a)
        root_2 = (-b + np.sqrt(b**2 - 4 * a * c)) / (2 * a)
    else:
        theta_EE = theta_EE_from_theta_D(det_alt, theta_D) * (np.pi / 180)
        a = 1
        b = 2 * const.R_Earth * np.sin(theta_EE)
        c = const.R_Earth**2 - (z + const.R_Earth)**2
        root_1 = (-b - np.sqrt(b**2 - 4 * a * c)) / (2 * a)
        root_2 = (-b + np.sqrt(b**2 - 4 * a * c)) / (2 * a)

    return max(np.array([root_1, root_2]))


def ellipse_focii(Ls, L_detector, electron_zenith, Cherenkov_angle):
    """Calculates the minor and major focii (in km) of the ellipse generated on the detection plane (assuming the plane normal to the shower incidence)
    given the distance of the emission point from the shower starting point (in km), the detection plane distance (in km),
    the electron zenith angle with respect to the shower axis (in degrees), and the Cherenkov angle (in degrees)

    Parameters
    ----------
    Ls : float
        distance emission point to shower starting point
    L_detector : float
        distance to detection plane
    electron_zenith : float
        angle to shower axis in degree
    Cherenkov_angle : float
        in degree

    Returns
    -------
    ArrayLike
        minor and major focii on detetcion plane in km
    """

    electron_zenith, Cherenkov_angle = (np.pi / 180) * electron_zenith, (np.pi / 180) * Cherenkov_angle

    minor_foci = (L_detector - Ls[None, :]) * (np.tan(Cherenkov_angle[None, :]) / np.cos(electron_zenith)) * np.sqrt(1 + np.sin(electron_zenith) / (1 - np.tan(electron_zenith)**2 * np.tan(Cherenkov_angle[None, :])**2))
    major_foci = minor_foci / np.sqrt(1 - np.tan(electron_zenith)**2 * np.tan(Cherenkov_angle[None, :])**2)

    return [minor_foci, major_foci]


def ellipse_coordinates(Ls, L_detector, electron_zenith, Cherenkov_angle, Cherenkov_azimuth):
    """Calculates the x and y coordinates of the ellipse generated on the detection plane (assuming the plane normal to the shower incidence)
    given the distance of the emission point from the shower starting point (in km), the detection plane distance (in km),
    the electron zenith angle with respect to the shower axis (in degrees), and the Cherenkov angle (in degrees)

    Parameters
    ----------
    Ls : float
        distance emission point to shower starting point
    L_detector : float
        distance to detection plane
    electron_zenith : [type]
        [description]
    Cherenkov_angle : float
        in degree
    Cherenkov_azimuth : float
        in degree

    Returns
    -------
    ArrayLike
        list of arrays with x,y coordinates of ellipse on the detection plane
    """

    # Calculating the major and minor focii of the generated ellipse
    focii = ellipse_focii(Ls, L_detector, electron_zenith, Cherenkov_angle)
    minor_foci, major_foci = focii[0], focii[1]

    electron_zenith, Cherenkov_angle = (np.pi / 180) * electron_zenith, (np.pi / 180) * Cherenkov_angle
    
    # Using the azimuth bin size difference to smear the coordinates
    dtheta = np.diff(Cherenkov_azimuth)[0]/2

    ellipse_x = (L_detector - Ls[None, :, None]) * np.tan(electron_zenith[:, :, None]) + (L_detector - Ls[None, :, None]) * ((np.sin(electron_zenith[:, :, None]) / np.cos(electron_zenith[:, :, None])**2) * np.tan(Cherenkov_angle[None, :, None])**2) / (1 - np.tan(electron_zenith[:, :, None])**2 * np.tan(Cherenkov_angle[None, :, None])**2) + major_foci[:, :, None] * np.cos(Cherenkov_azimuth)
    ellipse_y = minor_foci[:, :, None] * np.sin(Cherenkov_azimuth+dtheta*np.random.uniform(-1,1,minor_foci.shape).reshape(minor_foci.shape)[:, :, None])

    return [ellipse_x, ellipse_y]



# BSD 3-Clause License
#
# Copyright (c) 2021, Austin Cummings
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' Electron Functions

.. autosummary::
   :toctree:
   :recursive:

    Lorentz_factor
    particle_velocity
    gyroradius
    geomagnetic_deflection_correction
    Hillas_Energy_Distribution
    Hillas_Angular_Distribution
    Lafebre_Lateral_Distribution
    Lafebre_Lag_Time_Distribution
    Greisen_Longitudinal_Profile
    CORSIKA_interpolation

'''

import numpy as np
from scipy import interpolate
import scipy.stats as st

from . import io_func
from . import constants as const
from . import atmosphere as atm


def Lorentz_factor(energy):
    """Calculate the Lorentz factor of an electron with a given energy in MeV

    Parameters
    ----------
    energy : float
        e- energy in MeV

    Returns
    -------
    float
        Lorentz factor
    """
    gamma = energy / const.electron_mass

    return gamma


def particle_velocity(energy):
    """Calculate the speed of an electron relative to the speed of light with a given energy in MeV

    Parameters
    ----------
    energy : float
        e- energy in MeV

    Returns
    -------
    float
        speed of e- relative to c
    """
    # Calculate the particle's Lorentz factor
    particle_Lorentz_factor = Lorentz_factor(energy)

    beta = np.sqrt(1 - 1 / (particle_Lorentz_factor**2))

    return beta


def gyroradius(electron_energy, electron_zenith, electron_azimuth, magnetic_field_strength, magnetic_field_zenith):
    """alculates the gyroradius (in km) of an electron with a given energy (in MeV), zenith angle (with respect to trajectory: in degrees)
    and azimuth angle (in degrees) inside a magnetic field with a given strength (in uT) and zenith angle(in degrees)

    Parameters
    ----------
    electron_energy : float
        e- energy in MeV
    electron_zenith : float
        e- zenith in degree
    electron_azimuth : [type]
        e- azimuth in degree
    magnetic_field_strength : float
        in uT
    magnetic_field_zenith : float
        in degree

    Returns
    -------
    float
        gyro radius in km
    """
    # Calculates the normalized pointing array of the magnetic field, aligning the incident particle along the z axis
    magnetic_field_zenith = (np.pi / 180) * magnetic_field_zenith
    magnetic_field_direction = np.array([np.sin(magnetic_field_zenith), 0, np.cos(magnetic_field_zenith)])

    # Initializes the normalized velocity vector of the particle (aligned along the z axis)
    electron_zenith = (np.pi / 180) * electron_zenith
    particle_direction = np.array([np.cos(electron_azimuth[None, None, :]) * np.sin(electron_zenith[:, :, None]), np.sin(electron_azimuth[None, None, :]) * np.sin(electron_zenith[:, :, None]), np.cos(electron_zenith[:, :, None] * np.ones(len(electron_azimuth))[None, None, :])])

    # Calculating the Lorentz factor and particle relative velocity
    particle_Lorentz_factor, particle_relative_velocity = Lorentz_factor(electron_energy), particle_velocity(electron_energy)

    particle_gyroradius = particle_Lorentz_factor[None, :, None] * const.electron_mass * const.MeV_to_kg * particle_relative_velocity[None, :, None] * const.cl / (const.electron_charge * np.linalg.norm(np.cross(particle_direction, magnetic_field_strength * 1e-6 * magnetic_field_direction, axis=0), axis=0))

    return particle_gyroradius


def geomagnetic_deflection_correction(Ls, L_detector, electron_energy, electron_zenith, electron_azimuth, magnetic_field_strength, magnetic_field_zenith):
    """Calculates the deflection correction (in km) of an electron with a given energy (in MeV), zenith angle (with respect to trajectory: in degrees)
    and azimuth angle (in degrees) inside a magnetic field with a given strength (in uT) and zenith angle(in degrees)

    TODO: RECONSIDER THIS WHOLE CONCEPT...

    Parameters
    ----------
    Ls : [type]
        [description]
    L_detector : [type]
        [description]
    electron_energy : [type]
        in MeV
    electron_zenith : [type]
        in degree
    electron_azimuth : [type]
        in degree
    magnetic_field_strength : [type]
        in uT
    magnetic_field_zenith : [type]
        in degree

    Returns
    -------
    float
        deflection correction in km
    """
    # Calculate the gyroradius of the electrons
    particle_gyroradius = gyroradius(electron_energy, electron_zenith, electron_azimuth, magnetic_field_strength, magnetic_field_zenith)

    # Calculate the deflected angle inside the given distance bins and the corresponding projection on the detection plane
    dLs = np.append(np.diff(Ls), Ls[-1] - Ls[-2])
    added_distance = np.reshape(np.random.choice([-1, 1], size=particle_gyroradius.shape), particle_gyroradius.shape) * (L_detector - Ls[None, :, None]) * (dLs[None, :, None] / particle_gyroradius) / np.sqrt(1 - (dLs[None, :, None] / particle_gyroradius)**2)

    return added_distance


def Hillas_Energy_Distribution(electron_energy, shower_age):
    """Calculates the total fraction of electrons present at a given shower age above an energy (in MeV) as given by:
    A.M. Hillas. Air. J. Phys. G, 8:1461-1473, 1982

    Parameters
    ----------
    electron_energy : float
        in MeV
    shower_age : float

    Returns
    -------
    float
        e- fraction at shower age
    """
    E0 = np.where(shower_age < 0.4, 26, 44 - 17 * (shower_age - 1.46)**2)

    if isinstance(shower_age, np.ndarray):
        part1 = 0.89 * E0 - 1.2
        part2 = E0[:, None] + electron_energy
        part3 = 1 + (electron_energy * shower_age[:, None]) * 1e-4

        electron_fraction = ((part1[:, None] / part2)**shower_age[:, None]) * (part3)**(-2)
    else:
        part1 = 0.89 * E0 - 1.2
        part2 = E0 + electron_energy
        part3 = 1 + (electron_energy * shower_age) * 1e-4

        electron_fraction = ((part1 / part2)**shower_age) * (part3)**(-2)

    return electron_fraction


def Hillas_Angular_Distribution(E, s, us):
    """e- angular distribution (Hillas)

    Calculates (i) the corresponding angle (in degrees) of the scaled angular bin (u) and
    (ii) the fraction of electrons within the scaled angular bins
    for a given energy (in MeV) and shower age as given by: A.M. Hillas. Air. J. Phys. G, 8:1461-1473, 1982

    Parameters
    ----------
    E : [type]
        [description]
    s : [type]
        [description]
    us : [type]
        [description]

    Returns
    -------
    list of arrays
        angular distribution
    """
    # Calculate the midpoint of the angular bins
    lower_us, upper_us = us[:-1], us[1:]
    mid_us = (lower_us + upper_us) / 2

    # Calculate the corresponding angle of the given bins
    v = E / (1150 + 454 * np.log(s))
    average_w = 0.0054 * E * (1 + v) / (1 + 13 * v + 8.3 * v * v)
    w = mid_us[:, None] * average_w[None, :]
    theta = (180 / np.pi) * np.arccos(1 - w / (2 * (E[None, :] / 21)**2))

    # Calculate the fraction of electrons within each scaled angular bin
    A = np.where(E < 350, 0.777, 1.318)
    z0 = np.where(E < 350, 0.59, 0.37)
    z_values = [np.sqrt(lower_us), np.sqrt(upper_us)]
    lambda_ls = [np.where((E[None, :] < 350) & (z[:, None] < 0.59), 0.478, np.where((E[None, :] < 350) & (z[:, None] >= 0.59), 0.380, np.where((E[None, :] >= 350) & (z[:, None] < 0.37), 0.413, 0.380))) for z in z_values]

    # Must consider the bins in z which cross z0 and have different behavior
    [n_lower, n_upper] = [-2 * A[None, :] * np.exp((-z_values[i][:, None] + z0[None, :]) / lambda_ls[i]) * lambda_ls[i] * (lambda_ls[i] + z_values[i][:, None]) for i in (0, 1)]
    [n_z0_lower, n_z0_upper] = [-2 * A[None, :] * lambda_ls[i] * (lambda_ls[i] + z0[None, :]) for i in (0, 1)]

    dn = np.where((z_values[0][:, None] < z0[None, :]) & (z_values[1][:, None] >= z0[None, :]), n_upper - n_z0_upper + n_z0_lower - n_lower, n_upper - n_lower)

    # Do not return NaN values
    NaN_values = np.isnan(theta)
    theta[NaN_values] = 0
    dn[NaN_values] = 0

    return [theta, dn]


def Lafebre_Lateral_Distribution(electron_energies, radiation_lengths, zs, coordinates):
    """Lateral distribution Lafebre

    Randomly samples lateral offsets of electrons (in km) from the shower axis, given their primary energy (in MeV),
    their modified shower age (radiation length measured from shower maximum), their altitudes, and the desired shape of the output array
    as parameterized by: S. Lafebre, R. Engel, H. Falcke, J. Horandel, T. Huege, J. Kuijpers, and R. Ulrich. Astropart. Phys., 31:243-254, 2009

    Parameters
    ----------
    electron_energies : ArrayLike
        [description]
    radiation_lengths : ArrayLike
        [description]
    zs : ArrayLike
        [description]
    coordinates : ArrayLike
        [description]

    Returns
    -------
    [type]
        [description]
    """
    # Model is only parameterized within energy range 1MeV-1GeV
    modified_electron_energies = np.copy(electron_energies)
    modified_electron_energies[modified_electron_energies < 1] = 1
    modified_electron_energies[modified_electron_energies > 1000] = 1000

    # Model is only parameterized within shower age range -9-9
    radiation_lengths[radiation_lengths < -9] = -9
    radiation_lengths[radiation_lengths > 9] = 9

    # Model parameters
    logE = np.log(modified_electron_energies)
    x1 = 0.859 - 0.0461 * (logE**2) + 0.00428 * (logE**3)
    Lt = 0.0263 * radiation_lengths
    L0 = Lt + 1.34 + 0.160 * logE - 0.0404 * (logE**2) + 0.00276 * (logE**3)
    L1 = Lt - 4.33

    # Loading the desired shape of the samples
    zenith_shape, azimuth_shape = coordinates.shape[0], coordinates.shape[2]

    # Calculating Moliere Radius
    moliere = atm.moliere_radius(zs)

    # Randomly sample the lateral offsets using a scaled betaprime distribution
    sampled_offsets = np.transpose(np.array([moliere[i] * x1[i] * st.betaprime.rvs(L0[i], -L0[i] - L1[i], size=zenith_shape * azimuth_shape).reshape(zenith_shape, azimuth_shape) for i in range(0, len(modified_electron_energies))]), (1, 0, 2))

    return sampled_offsets


def Lafebre_Lag_Time_Distribution(electron_energies, radiation_lengths, zs, coordinates):
    """Randomly samples lag times of electrons (in s) from an imaginary particle travelling along the shower axis at the speed of light, given their primary energy (in MeV),
    their modified shower age (radiation length measured from shower maximum), their altitudes, and the desired shape of the output array
    as parameterized by: S. Lafebre, R. Engel, H. Falcke, J. Horandel, T. Huege, J. Kuijpers, and R. Ulrich. Astropart. Phys., 31:243-254, 2009

    Parameters
    ----------
    electron_energies : ArrayLike
        e- energies in MeV
    radiation_lengths : ArrayLike

    zs : [type]
        [description]
    coordinates : ArrayLike
        [description]

    Returns
    -------
    ArrayLike
        lag times of electrons in s
    """
    # Model is only parameterized within energy range 1MeV-1GeV
    modified_electron_energies = np.copy(electron_energies)
    modified_electron_energies[modified_electron_energies < 1] = 1
    modified_electron_energies[modified_electron_energies > 1000] = 1000

    # Model is only parameterized within shower age range -4.5 to 9
    # lower limit adjusted to respected domains scaled betaprime distribution
    radiation_lengths[radiation_lengths < -4.45] = -4.45
    radiation_lengths[radiation_lengths > 9] = 9

    # Model parameters
    logE = np.log(modified_electron_energies)
    tau1 = np.exp(-2.71 + 0.0823 * logE - 0.114 * (logE**2))
    L0 = 1.70 + 0.160 * radiation_lengths - 0.142 * logE
    L1 = -3.21

    # Loading the desired shape of the samples
    zenith_shape, azimuth_shape = coordinates.shape[0], coordinates.shape[2]

    # Calculating Moliere Radius
    moliere = atm.moliere_radius(zs)

    # Randomly sample the lateral offsets using a scaled betaprime distribution
    sampled_lag_times = (1 / const.cl) * np.transpose(np.array([moliere[i] * tau1[i] * st.betaprime.rvs(L0[i], -L0[i] - L1, size=zenith_shape * azimuth_shape).reshape(zenith_shape, azimuth_shape) for i in range(0, len(modified_electron_energies))]), (1, 0, 2))

    return sampled_lag_times


def Greisen_Longitudinal_Profile(energy, slant_depth):
    """Calculates the number of charged particles for a given slant depth (in g/cm^2) and a given shower energy (in eV)
    as parameterized by Greisen, found in: Cosmic Rays and Particle Physics: 2nd Edition.

    Parameters
    ----------
    energy : float
        in eV
    slant_depth : float
        in g/cm^2

    Returns
    -------
    [type]
        [description]
    """
    # Convert the slant depth to radiation lengths
    radiation_length = slant_depth / 36.7

    # Convert the radiation lengths to shower age
    shower_age = 3 * radiation_length / (radiation_length + 2 * np.log(energy / 87e6))

    # Calculate the number of charged particles and filter out and NaN values
    charged_particles = (0.31 / ((np.log(1e17 / 87e6))**0.5)) * np.exp(radiation_length * (1 - 1.5 * np.log(shower_age)))
    charged_particles[np.isnan(charged_particles)] = 0

    return charged_particles


def CORSIKA_interpolation(energy, slant_depth, filename, simulated_energy):
    """Calculates the number of charged particles for a given slant depth (in g/cm^2) and a given shower energy (in eV)
    interpolating from a given charged particle longitudinal profile generated by CORSIKA and scaled linearly

    Parameters
    ----------
    energy : float
        shower energy in eV
    slant_depth : float
        in g/cm^2
    filename : str
        name of CORSIKA file
    simulated_energy : float
        shower energy used for CORSIKA simualtion in eV

    Returns
    -------
    ArrayLike
        charged particle profile scaled to shower energy
    """
    # Load in the CORSIKA showers
    CORSIKA_showers = io_func.CORSIKA_Shower_Reader(filename)

    # Calculate the average of these showers
    average_shower = io_func.average_CORSIKA_profile(CORSIKA_showers)
    del CORSIKA_showers

    # Create an interpolation object of the average shower
    charged_particles_from_slant_depth = interpolate.interp1d(average_shower[0], average_shower[1], fill_value='extrapolate')
    del average_shower

    # Calculate the number of charged particles and filter out NaN and negative values
    charged_particles = (energy / simulated_energy) * charged_particles_from_slant_depth(slant_depth)
    charged_particles[np.isnan(charged_particles)] = 0
    charged_particles[charged_particles < 0] = 0

    return charged_particles




# The Clear BSD License
#
# Copyright (c) 2021 Austin Cummings
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#
#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#      * Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.
#
# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

""" easchersim

.. autosummary::
   :toctree:
   :recursive:

   cli
   run

"""

import argparse
import sys

import numpy as np

from . import io_func
from . import geometry as geo
from . import constants as const
from . import spatial_cherenkov_profile as sp_cher_prof


def cli():
    """Command line interface

    Returns
    -------
    parser Obj
        command line arguments
    """
    parser = argparse.ArgumentParser(prog='easchersim', usage='%(prog)s [OPTIONS] COMMAND [ARGS]')
    parser._optionals.title = "OPTIONS"
    subparsers = parser.add_subparsers(title='COMMANDS', dest='COMMAND', required=True)
    # run commad
    run_par = subparsers.add_parser('run',
                                    help='run easchersim for settings from configuration file',
                                    prog='easchersim run',
                                    usage='%(prog)s [OPTIONS] CONFIGFILE'
                                    )
    run_par._optionals.title = "OPTIONS"
    run_par.add_argument('filename',
                         help=argparse.SUPPRESS
                         )
    run_par.add_argument('-e', '--energy',
                         type=float,
                         dest='energy',
                         help='Set shower energy in eV'
                         )
    run_par.add_argument('-a', '--angle',
                         nargs=2,
                         metavar=("ANGLE", "GEOMETRY-DEF"),
                         dest='angle',
                         help='Set shower angle and geometry definition (viewing/emergence)'
                         )
    run_par.add_argument('-d', '--dist-first-int',
                         type=float,
                         dest='distFirstInt',
                         help='Set distance to first interaction'
                         )
    run_par.set_defaults(func=run)

    # create the configuration file
    make_cfg = subparsers.add_parser('make-config',
                                     help='Produce configuration file from default \
                                     and/or given parameters.',
                                     prog='easchersim make-config',
                                     usage='%(prog)s [OPTIONS] CONFIGFILE')
    make_cfg._optionals.title = "OPTIONS"
    make_cfg.add_argument("filename",
                          default="easchersim.cfg",
                          help=argparse.SUPPRESS)

    make_cfg.add_argument('--det-alt',
                          type=float,
                          default=33,
                          dest='det_alt',
                          help='Set detection plane altitude in km'
                          )
    make_cfg.add_argument('-e', '--energy',
                          type=float,
                          default=1.e17,
                          dest='energy',
                          help='Set shower energy in eV'
                          )
    make_cfg.add_argument('-a', '--angle',
                          nargs=2,
                          default=[0., 'viewing'],
                          metavar=("ANGLE", "GEOMETRY-DEF"),
                          dest='angle',
                          help='Set shower angle and geometry definition (viewing/emergence)'
                          )
    make_cfg.add_argument('-d', '--dist-first-int',
                          type=float,
                          default=0,
                          dest='distFirstInt',
                          help='Set distance to first interaction'
                          )
    make_cfg.add_argument('-s', '--scattering',
                          type=str,
                          choices=('yes', 'no', 'on', 'off'),
                          default='yes',
                          dest='atm_scatter',
                          help='Turn atmospheric scattering on/off'
                          )
    make_cfg.add_argument('-m', '--use-mag-field',
                          type=str,
                          choices=('yes', 'no', 'on', 'off'),
                          default='on',
                          dest='mag_field',
                          help='Turn magnetic field on/off'
                          )
    make_cfg.add_argument('-g', '--use-greisen',
                          type=str,
                          choices=('yes', 'no', 'on', 'off'),
                          default='no',
                          dest='useGreisen',
                          help='Use Greisen parametrization for longitudianal profile \
                                instead of CORSIKA file'
                          )
    make_cfg.add_argument('-o', '--output',
                          type=str,
                          default='',
                          dest='output_name',
                          help='Name of the output file'
                          )
    make_cfg.add_argument('-p', '--plots',
                          type=str,
                          choices=('none', 'basic', 'general', 'all'),
                          default='general',
                          dest='plots',
                          help='Set the type of plots to be created'
                          )
    make_cfg.set_defaults(func=io_func.create_config)
    return parser.parse_args()


def run(user_args):
    """Perform easchersim simulation

    Main application of easchersim, performs the simulation based on settings given config file

    Parameters
    ----------
    user_args : argparse Obj
        command line arguments

    Examples
    --------
    Command line usage of the run command:
    `easchersim run my_config_file.ini`
    """
    cfg = io_func.read_config(user_args.filename)

    # Check provided configuration
    detector_altitude = cfg.getfloat('detector', 'altitude')
    if user_args.angle is None:
        geo_view = cfg.getlist('shower', 'angle')[1]
        theta = float(cfg.getlist('shower', 'angle')[0])
    else:
        geo_view = user_args.angle[1]
        theta = float(user_args.angle[0])

    if geo_view == "emergence":
        theta_D = geo.theta_D_from_theta_EE(detector_altitude, theta)
    elif geo_view == "viewing":
        theta_D = theta
    else:
        io_func.flog(f'''[bold blue]WARN:[/] You entered an unkown shower geometry ({geo_view})! \n'''
                     '''Allowed options are\n'''
                     '''   - emergence for emergence angle shower definition\n'''
                     '''   - viewing for detector viewing angle shower definition\n'''
                     '''Please change your configuration accordingly.\n''')
        sys.exit(-1)

    if user_args.distFirstInt is None:
        L_particle = cfg.getfloat('shower', 'distFistInteraction')
    else:
        L_particle = float(user_args.distFirstInt)
    ## Calculating the path lengths to the detector and the top of the atmosphere
    [L_detector, L_atmosphere] = [geo.L_from_z(z, theta_D, detector_altitude)
                                  for z in [detector_altitude, const.z_max]]

    ## Disallowing for unobservable trajectories
    if np.isnan(L_detector) or np.isnan(L_atmosphere) or L_particle > min([L_detector, L_atmosphere]):
        io_func.flog('''[bold red] ERROR: [/]EAS begins past either the detector or the atmosphere and is unobservable.''')
        # change to try next shower in the future
        sys.exit(-1)
    if user_args.energy is None:
        energy = cfg.getfloat('shower', 'energy')
    else:
        energy = float(user_args.energy)

    sp_cher_prof.cherenkov_from_particle_profile(detector_altitude,
                                                 energy, theta_D,
                                                 L_particle, L_detector, L_atmosphere,
                                                 cfg['atmosphere'], cfg['magField'],
                                                 cfg['runSettings'],
                                                 cfg.get('output', 'plots')
                                                 )


def main():
    io_func.get_console(log_time=False, log_path=False).rule("[bold blue]EASCherSim")
    args = cli()
    args.func(args)


if __name__ == "__main__":
    main()



# BSD 3-Clause License
#
# Copyright (c) 2021, Austin Cummings
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np

################################################################
#Fitting Cherenkov Spatial Profile

#Exponential Profile
def custom_fit(x, x0, Nmax, scale):
	y = np.piecewise(x, [np.abs(x) < x0, np.abs(x) >= x0],
					 [lambda x: Nmax, lambda x: Nmax*np.exp(-(np.abs(x)-np.abs(x0))/scale)])
	return y

#Gaussian Profile
def gaussian_fit(x, x0, Nmax):
	y = np.piecewise(x, [np.abs(x) < x0, np.abs(x) >= x0],
					 [lambda x: Nmax, lambda x: Nmax*np.exp(-((np.abs(x)-x0)**2)/(2*x0**2))])
	return y

def mod_gauss_fit(x, x0, x1, Nmax, scale):
	y = np.piecewise(x, [np.abs(x) < x0, np.abs(x) >= x0],
					 [lambda x: np.log(Nmax), lambda x: np.log(Nmax*np.exp(-((np.abs(x)-x0)/x1)**scale))])
	return y

#Polynomial Profile (edges go down as 1/x^n)
def log_fit(x, x0, Nmax, scale):
	y = np.piecewise(x, [np.abs(x) < x0, np.abs(x) >= x0],
					 [lambda x: Nmax, lambda x: Nmax*(x0/np.abs(x))**scale])
	return y

def better_fit(x, x0, x1, x2, A, scale):
	y = np.piecewise(x, [np.abs(x) < x0, (np.abs(x) >= x0)&(np.abs(x) <= x1), np.abs(x)>x1],
					 [lambda x: A, lambda x: A+scale*np.log((x0/x)), lambda x: A+scale*np.log(x0/x1)-(x-x1)/x2])
	return y

def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx

#Gets the weighted percentiles of data
def weighted_percentile(data, percents, weights=None):
	''' percents in units of 1%
		weights specifies the frequency (count) of data.
	'''
	if weights is None:
		return np.percentile(data, percents)
	ind=np.argsort(data)
	d=data[ind]
	w=weights[ind]
	p=1.*w.cumsum()/w.sum()*100
	y=np.interp(percents, p, d)
	return y


# BSD 3-Clause License
#
# Copyright (c) 2021, Austin Cummings
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# TODO: make Showerfile actually configurable, requires calculation of X_0 and energy of shower file
# TODO: Re-write onfiguration parser

''' IO Functions

.. autosummary::
   :toctree:
   :recursive:

   CORSIKA_Shower_Reader
   average_CORSIKA_profile
   read_config

'''

from rich.console import Console
from rich.table import Column
from rich.progress import Progress, BarColumn, TextColumn, TimeElapsedColumn

import configparser as cp
import numpy as np
import io

'''
root_found = True
try:
    from ROOT import TFile, TTree, TH1D, vector
except ImportError:
    print("The root application is not available! If you want to have root output please install.")
    root_found = False
'''
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files


class IO:
    """root IO class
    """
    def __init__(self, file_name='CherSim.root'):
        self.__vec_time = vector("TH1D")()
        self.__vec_ang = vector("TH1D")()
        self.create_root_file(file_name)

    def create_root_file(self, filename):
        self.__file = TFile(filename, 'recreate')
        self.__tcher_ph = TTree("cherPhProp", "Cherenkov Photon Properties")
        self.__tshower = TTree("showerProp", "Shower Properties")

        # create branches for tshower
        self.__eng = np.zeros(1, dtype=float)
        self.__zen = np.zeros(1, dtype=float)
        self.__startalt = np.zeros(1, dtype=float)

        self.__tshower.Branch("energy", self.__eng, 'eng/D')
        self.__tshower.Branch("zenith", self.__zen, 'zen/D')
        self.__tshower.Branch("startalt", self.__startalt, 'startalt/D')

        # define histograms
        histo_w = TH1D()
        histo_r = TH1D()
        histo_t_off = TH1D()
        histo_ang_off = TH1D()

        # set branch for histograms
        self.__tcher_ph.Branch("wavelength", 'TH1D', histo_w)
        self.__tcher_ph.Branch("distance", 'TH1D', histo_r)
        self.__tcher_ph.Branch("time_offset", 'TH1D', histo_t_off)
        self.__tcher_ph.Branch("angle_offset", 'TH1D', histo_ang_off)
        self.__tcher_ph.Branch("time_dist", self.__vec_time)
        self.__tcher_ph.Branch("angle_dist", self.__vec_ang)

    def write_shower_prop(self, energy, zenith, start):
        print("Write Shower Prop")

        # add shower properties
        self.__eng[0] = energy
        self.__zen[0] = zenith
        self.__startalt[0] = start
        self.__tshower.Fill()

    def write_cher_ph_prop(self, cher_ph):
        print("Write Photon Prop")

        num_wl_bin = len(cher_ph["w_grid"]) - 1
        num_dist_bin = len(cher_ph["r_bins"]) - 1
        num_time_bin = len(cher_ph["time_bins"]) - 1
        num_ang_bin = len(cher_ph["angle_bins"]) - 1

        # define histograms
        histo_w = TH1D("wl", "wavelength", num_wl_bin, cher_ph["w_grid"])
        histo_r = TH1D("r", "distance", num_dist_bin, cher_ph["r_bins"])
        histo_t_off = TH1D("t_off", "time offset",
                           num_dist_bin, cher_ph["r_bins"])
        histo_ang_off = TH1D("ang_off", "angle offset",
                             num_dist_bin, cher_ph["r_bins"])
        histo_t = [TH1D("t_dist_" + str(i), "time_dist_" + str(i),
                        num_time_bin, cher_ph["time_bins"]) for i in range(num_dist_bin)]
        histo_ang = [TH1D("ang_dist_" + str(i), "angle_dist_" + str(i),
                          num_ang_bin, cher_ph["angle_bins"]) for i in range(num_dist_bin)]

        # fill histograms
        for wl_bin, counts in enumerate(cher_ph["w_counts"]):
            histo_w.SetBinContent(wl_bin + 1, counts)
        for r_bin, counts in enumerate(cher_ph["r_counts"]):
            histo_r.SetBinContent(r_bin + 1, counts)
            histo_t_off.SetBinContent(r_bin + 1, cher_ph["lower_times"][r_bin])
            histo_ang_off.SetBinContent(r_bin + 1, cher_ph["lower_angles"][r_bin])
            for t_bin, counts_t in enumerate(cher_ph["time_counts"][r_bin]):
                histo_t[r_bin].SetBinContent(t_bin + 1, counts_t)
            for ang_bin, counts_ang in enumerate(cher_ph["angle_counts"][r_bin]):
                histo_ang[r_bin].SetBinContent(ang_bin + 1, counts_ang)

        self.__tcher_ph.SetBranchAddress("wavelength", histo_w)
        self.__tcher_ph.SetBranchAddress("distance", histo_r)
        self.__tcher_ph.SetBranchAddress("angle_offset", histo_ang_off)
        self.__tcher_ph.SetBranchAddress("time_offset", histo_t_off)
        self.__vec_time.assign(histo_t)
        self.__vec_ang.assign(histo_ang)
        self.__tcher_ph.Fill()

    def close_root(self):
        self.__tshower.Write("", TFile.kOverwrite)
        self.__tcher_ph.Write("", TFile.kOverwrite)
        self.__file.Close()


def is_float(s):
    """Determines if an input value is a float

    Parameters
    ----------
    s : str
        input string

    Returns
    -------
    Bollean
        True: if string is float, False otherwise
    """

    try:
        float(s)
        return True
    except ValueError:
        return False


def CORSIKA_Shower_Reader(filename):
    """Reads in longitudinal charged particle profiles (.long files) generated by CORSIKA simulations

    Parameters
    ----------
    filename : str
        CORSIKA .long file path

    Returns
    -------
    ArrayLike
        longitudial chaged particle profile of showers array of lists (depth, num electrons)
    """
    # Initialize an array to contain all the simulated showers and arrays to contain the depths (g/cm^2) and particle content for each shower
    showers = []
    depths, electrons = [], []

    with open(filename) as f:
        # Flag for whether to output the lines which follow, initialize as false
        readout_flag = False

        # Loop through the given file line by line
        for line in f:
            if readout_flag:

                # Split the line on the " " delimiter
                data = line.split()

                # If the first value is a float, can continue
                if is_float(data[0]):
                    # Append the depth to the depth array, and the sum of electrons and positrons to the electron array
                    # If other particle types are desired:
                    # [0]: depths, [1]: gammas, [2]: electrons, [3]: positrons, [4]: anti-muons, [5]: muons, [6]: hadrons, [7]: charged particles, [8]: nuclei
                    depths.append(float(data[0]))
                    electrons.append(float(data[2]) + float(data[3]))

            # Determine if charged particle info follows.
            # CORSIKA outputs charged particle profiles, followed by energy information, so flag must switch every instance of header
            if "ELECTRONS" in line:
                # Append shower info to container array
                if len(depths) > 0:
                    showers.append([depths, electrons])

                readout_flag = not readout_flag
                # Reinitialize arrays
                depths, electrons = [], []

    showers = np.array(showers)

    return showers


def average_CORSIKA_profile(showers):
    """Calculates the average charged particle longitudinal profile from the CORSIKA data

    Parameters
    ----------
    showers : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    average_shower = np.average(showers, axis=0)

    return average_shower


def getargs():
    """Read of command line arguments

    Returns
    -------
    Namespace
        returns commandline arguments
    """
    cli = argparse.ArgumentParser()
    cli.add_argument(
        "--config",
        type=str,
        help="File containing simulation configurations",
        required=True
    )

    args = cli.parse_args()
    return args


def read_config(cfg_file):
    """Read configuration file.

    Parameters
    ----------
    cfg_file : str
        name of User config file

    Returns
    -------
    ConfigParser
        Settings for simulation (defualt & user specified)
    """
    cfg = cp.ConfigParser(allow_no_value=True,
                          converters={'list': lambda x: [i.strip() for i in x.split(',')]})
    cfg.optionxform = lambda option: option
    # read default settings
    default_cfg = files("EASCherSim.data")/"default_conf.ini"
    cfg.read(default_cfg)

    # read user settings (overwrittes defaults)
    cfg.read(cfg_file)

    return cfg


def create_config(user_args):
    """Creates example configuration file

    Values can be specified by user via command line all not specified parameter will be default

    Parameters
    ----------
    user_args : namespace
        User inputs parsed by argparse
    """
    cfg = cp.ConfigParser()
    cfg.optionxform = lambda option: option
    cfg['shower'] = {'energy': user_args.energy,
                     'angle': f"""{user_args.angle[0]}, {user_args.angle[1]}""",
                     'distFistInteraction': user_args.distFirstInt}
    cfg['detector'] = {'altitude': user_args.det_alt}
    cfg['atmosphere'] = {'scattering': user_args.atm_scatter}
    cfg['magField'] = {'useField': user_args.mag_field}
    cfg['runSettings'] = {'useGreisenParametrization': user_args.useGreisen}
    cfg['output'] = {'name': user_args.output_name}
    cfg['output'] = {'plots': user_args.plots}

    with open(user_args.filename, 'w') as cfg_file:
        cfg.write(cfg_file)


def dump_default_conf():
    """Print default configuration to screen
    """
    cfg = cp.ConfigParser()
    # read default settings
    default_cfg = files("EASCherSim.data")/"default_conf.ini"
    cfg.read(default_cfg)
    with io.StringIO() as str_stream:
        cfg.write(str_stream)
        print(str_stream.getvalue())


def get_console(**kargs):
    if len(kargs) != 0:
        console = Console(**kargs, width=100)
    else:
        console = Console(width=100, log_path=False)
    return console


def flog(*args):
    get_console().log(*args)


def get_progress():
    text_col = TextColumn("[bold red]{task.description}", table_column=Column(ratio=1))
    bar_col = BarColumn(bar_width=40, table_column=Column(ratio=4))

    progress = Progress(text_col, bar_col, "[progress.percentage]{task.percentage:>3.0f}%",
                        TimeElapsedColumn(), console=get_console(),)
    return progress



# The Clear BSD License
#
# Copyright (c) 2021 Austin Cummings
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#
#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#      * Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.
#
# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

r"""EASCherSim - Simulate optical Cherenkov emission of Extensive Air Showers.
Python package and library.

***********
Application
***********

Application

.. autosummary::
   :toctree:

   EASCherSim.eas_cher_sim


******************
Simulation Modules
******************

Scientific modules

.. autosummary::
   :toctree:
   :nosignatures:

   EASCherSim.atmosphere
   EASCherSim.cher_func
   EASCherSim.electron_func
   EASCherSim.spatial_cherenkov_profile

*****************
Utility Functions
*****************

Supporting classes and functions for constants, IO operation, plotting etc.

.. autosummary::
   :toctree:
   :nosignatures:

   EASCherSim.constants
   EASCherSim.fit_func
   EASCherSim.geometry
   EASCherSim.io_func
   EASCherSim.plot

"""

__all__ = [
    "atmosphere",
    "cher_func",
    "constants",
    "electron_func",
    "fit_func",
    "geometry",
    "io_func",
    "spatial_cherenkov_profile",
    "plot"
]

from . import atmosphere, cher_func, constants, electron_func, fit_func, geometry, io_func, spatial_cherenkov_profile, plot



# BSD 3-Clause License
#
# Copyright (c) 2021, Austin Cummings
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


''' Atmospheric Functions

.. autosummary::
   :toctree:
   :recursive:

   atmospheric_extinction_coefficient
   atmos_density
   moliere_radius
   number_density
   index_of_refraction
   rayleigh_extinction_coefficient
   optical_depth
   slant_X_from_L_table

'''

import numpy as np
from scipy.interpolate import griddata
from scipy import integrate

from . import constants as const
from . import geometry as geo


def atmospheric_extinction_coefficient(file):
    r"""Loads in 2D array of atmospheric extinction coefficients from predefined file (over altitude and wavelength).

    Base implementation uses the results of L. Elterman. Number Tech. Rep. AFCRL-68-0153. 1968.

    Parameters
    ----------
        file: str
            conatining atmospheric extinction coefficients over altitude and wavelength

    Returns
    -------
        list
            wavelength in nm, altitudes in km, coefficients in m^-1
    """

    # Load wavelengths in nm
    wavelengths = np.loadtxt(file, delimiter=',', max_rows=1)

    # Load altitudes in km
    altitudes = np.loadtxt(file, delimiter=',', skiprows=1, max_rows=1)

    # Load in extinction coefficients in km^-1
    coefficients = np.loadtxt(file, delimiter=',', skiprows=2)

    # Remove any 0 values from array
    coefficients[coefficients == 0.0] = 1e-10

    # Convert to m^-1 and take the logarithm to smooth the data
    coefficients = np.log(coefficients / 1000)

    return [wavelengths, altitudes, coefficients]


def atmos_density(z):
    """Calculates the atmospheric density (in g/cm^3) of the atmosphere using Linsley's model for input altitude(s) (in km)

    Parameters
    ----------
        z: float
            altitude in km

    Returns
    -------
        float
            atm density in g/cm^3
    """

    # Find the appropriate atmospheric layer for the given altitude(s)
    if isinstance(z, np.ndarray):
        layers = np.argmax(np.less(z[:, None], const.atmos_layer_boundaries[None, :]), axis=1)

    else:
        layers = np.argmax(np.less(z, const.atmos_layer_boundaries))

    # Calculate the atmospheric density
    density = np.where(layers < 4, (const.atmos_b[layers] / const.atmos_c[layers]) * np.exp(-z * 1e5 / const.atmos_c[layers]), 1 / const.atmos_c[layers])

    # Above the maximum altitude in the model, the density is 0
    density[z > const.z_max] = 0

    return density


def moliere_radius(z):
    """Calculates the Moliere radius.

    Calculates the Moliere radius (the radius within which 90% of the shower energy deposition is contained: in km) at a given altitude (in km)
    as parameterized by: Dova, M. T., Epele, L. N., & Mariazzi, A. G. 2003, Astropart. Phys., 18, 351

    Parameters
    ----------
        z: float
            altitude in km

    Returns
    -------
        float
            Moliere radius in km
    """

    # Calculate the atmospheric density
    atmospheric_density = atmos_density(z)

    # Calculate the Moliere radius
    moliere_radius = 1e-5 * 9.6 / atmospheric_density

    return moliere_radius


def number_density(z):
    """Calculates the particle number density (in particles/cm^3) of the atmosphere for input altitude(s) (in km)

    Parameters
    ----------
        z: float
            altitude in km

    Returns
    -------
        float
            particle number density (in particles/cm^3)
    """

    return atmos_density(z) * (const.N_A / const.mm_air) * 1e-3


def index_of_refraction(z, wavelength):
    """Calculates the index of refraction.

    Calculates the index of refraction for input altitude(s) (in km) and wavelength(s) (in nm)
    Baseline approximation of the index of refraction, as given in:
    Handbook of Chemistry and Physics, 67th Edition, R.C. Weast ed. (The Chemical Rubber Co., Cleveland, 1986) E373
    
    Wavelength correction to the index of refraction as given in K. Bernlohr, Astropart. Phys., 30:149-158, 2008.

    Parameters
    ----------
        z: float
            altitude in km
        wavelength: float
            wavelength in nm

    Returns
    -------
        float
            index of refraction
    """

    n1 = 1 + (0.283e-3) * atmos_density(z) / atmos_density(0)

    if isinstance(z, np.ndarray) and isinstance(wavelength, np.ndarray):
        n = 1 + (n1[:, None] - 1) * (0.967 + 0.033 * (400 / wavelength[None, :])**2.5)
    else:
        n = 1 + (n1 - 1) * (0.967 + 0.033 * (400 / wavelength)**2.5)

    return n


def rayleigh_extinction_coefficient(z, wavelength):
    """ Calculates the Rayleigh extinction coefficient.

    Calculates the Rayleigh extinction coefficient (in m^-1) given input altitude(s) (in km) and wavelength(s) (in nm) a, 375-384
    King correction factor (Fk) to account for depolarization. Constant results from:
    R. Thalman, K. J. Zarzana, M. A. Tolbert, and R. Volkamer., Air. J. Quant. Spectrosc. Radiat. Transf., (147):171-177, 2014.
    
    Lord Rayleigh, The London, Edinburgh, and Dublin Philosophical Magazine and Journal of Science, 47:287

    Parameters
    ----------
        z: float
            altitude in km
        wavelength: float
            wavelength in nm

    Returns
    -------
        float
            Rayleigh extinction coefficient (in m^-1)
    """

    n2 = index_of_refraction(z, wavelength)**2
    N = number_density(z) * 1e6

    Fk = 1.0608

    alpha_Rayleigh = 24 * np.pi**3 * (((n2 - 1) / (n2 + 2))**2) * Fk * (1 / ((wavelength * 1e-9)**4)) * (1 / N[:, None])

    return alpha_Rayleigh


def optical_depth(data_wavelength, data_altitude, data_extinction_coefficients, wavelength_bins, L_bins, det_alt, theta_D, Rayleigh_flag):

    """Calculates the optical depth.

    Calculates the optical depth (tau: unitless) between emission and detection given distance along trajectory (in km), wavelength (in nm)
    and detector viewing angle (in degrees). Light is attenuated by e^(-tau(z,w))
    Takes in a boolean flag for whether to consider attenuation above 50km altitudes
    Ensure that the input wavelength bins do not exceed the bounds of the data (270nm-4000nm for Elterman 1968)

    Parameters
    ----------
    data_wavelength : [type]
        [description]
    data_altitude : [type]
        [description]
    data_extinction_coefficients : [type]
        [description]
    wavelength_bins : [type]
        [description]
    L_bins : [type]
        [description]
    det_alt : float
        detector altitude in km
    theta_D : [type]
        [description]
    Rayleigh_flag : bool
        true if consider attenuation above 50km otherwise false

    Returns
    -------
    ArrayLike
        ptical depth (tau: unitless)
    """

    # Flip the input distance bins because the optical depth must be calculated FROM the detector
    L_bins = L_bins[::-1]

    data_wavelength, data_altitude = np.meshgrid(data_wavelength, data_altitude)
    data_extinction_coefficients = data_extinction_coefficients.T

    # Calculate the altitudes of the distances along the trajectory
    altitude_bins = geo.z_from_L(L_bins, theta_D, det_alt)

    wavelength_bins, altitude_bins = np.meshgrid(wavelength_bins, altitude_bins)

    # 2D interpolation of the extinction coefficients
    # If Rayleigh_flag == True, above 50km, consider ONLY Rayleigh scattering
    # If False, no scattering above 50km
    if Rayleigh_flag:
        extinction_coefficients = np.where(altitude_bins > 50, rayleigh_extinction_coefficient(altitude_bins[:, 0], wavelength_bins[0]), np.exp(griddata((data_wavelength.flatten(), data_altitude.flatten()), data_extinction_coefficients.flatten(), (wavelength_bins, altitude_bins), method='cubic', fill_value=0)))
        extinction_coefficients[altitude_bins > const.alt_max] = 0
    else:
        extinction_coefficients = np.exp(griddata((data_wavelength.flatten(), data_altitude.flatten()), data_extinction_coefficients.flatten(), (wavelength_bins, altitude_bins), method='cubic', fill_value=0))
        extinction_coefficients[altitude_bins > 50] = 0

    # Cumulatively integrate the extinction coefficients along the trajectory
    optical_depth = -integrate.cumtrapz(extinction_coefficients.T, L_bins * 1e3, axis=1, initial=0)
    optical_depth = np.flip(optical_depth, axis=1)

    return optical_depth


def slant_X_from_L_table(L_bins, det_alt, theta_D):
    """Calculates the grammage profile (in g/cm^2) along the trajectory, given the detector altitude (in km)
    and the detector viewing angle (in degrees)

    Parameters
    ----------
    L_bins : [type]
        [description]
    det_alt : [type]
        [description]
    theta_D : [type]
        [description]

    Returns
    -------
    list
        grammage profile in g/cm^2
    """

    # Calculating a finer spacing of the distance bins along the trajectory
    L_bins = np.linspace(min(L_bins), max(L_bins), 10000)

    # Calculating the altitudes and densities of the positions along the trajectory
    altitude_bins = geo.z_from_L(L_bins, theta_D, det_alt)

    density = atmos_density(altitude_bins)

    # Integrating the grammage along the trajectory
    grammage = integrate.cumtrapz(density, L_bins * 1e5, initial=0)

    return [L_bins, grammage]



# BSD 3-Clause License
#
# Copyright (c) 2021, Austin Cummings
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Cherenkov functions

.. autosummary::
   :toctree:
   :recursive:

   Cherenkov_threshold
   Cherenkov_angle
   Cherenkov_photons

'''

import numpy as np
from . import constants as const
from . import atmosphere as atm


def Cherenkov_threshold(z, wavelength):
    """Calculates the threshold energy for Cherenkov

    Calculates the threshold energy for Cherenkov production by electrons at a given altitude (in km) and wavelength (in nm)
    Cherenkov theory originally given in: Cherenkov, P. A. (1934). "Visible emission of clean liquids by action of gamma radiation".
    Doklady Akademii Nauk SSSR. 2: 451.

    Parameters
    ----------
    z : float
        altitude in km
    wavelength : float
        wavelength in nm

    Returns
    -------
    float
        threshold energy in MeV
    """
    threshold_energy = const.electron_mass * np.sqrt(1 / (1 - 1 / (atm.index_of_refraction(z, wavelength)**2)))

    return threshold_energy


def Cherenkov_angle(z, wavelength, beta):
    """ Calculates Cherenkov angle.

    Calculates the angle at which the Cherenkov emission is emitted (in degrees) at a given altitude (in km), wavelength (in nm), and beta value (v/c)
    Cherenkov theory originally given in: Cherenkov, P. A. (1934). "Visible emission of clean liquids by action of gamma radiation".
    Doklady Akademii Nauk SSSR. 2: 451.

    Parameters
    ----------
    z : float
        altitude in km
    wavelength : float
        wavelength in nm
    beta : float
        beta = v/c

    Returns
    -------
    float
        Cherenkov angle in degree
    """
    theta_Cherenkov = (180 / np.pi) * np.arccos(1 / (beta * atm.index_of_refraction(z, wavelength)))

    return theta_Cherenkov


def Cherenkov_photons(z, wavelength, beta):
    """Calculates number of Cherenkov photons

    Calculate the number of Cherenkov photons produced by a single charged particle per unit wavelength and distance (m^-2)
    for a given altitude (in km), wavelength (in nm), and beta value (v/c)
    Theory given in: Tamm, I.E.; Frank, I.M. (1937), Dokl. Akad. Nauk SSSR, 14: 107

    Parameters
    ----------
    z : float
        altitude in km
    wavelength : float
        wavelength in nm
    beta : float
        beta = v/c

    Returns
    -------
    float
        [description]
    """
    n1 = atm.index_of_refraction(z, wavelength)

    if isinstance(beta, np.ndarray) and isinstance(wavelength, np.ndarray):
        dNdLdw = ((2 * np.pi * const.alpha_fsc) / (wavelength[None, None, :] * 1e-9)**2) * (1 - 1 / (beta[:, :, None] * n1[:, None, :])**2)
    else:
        dNdLdw = ((2 * np.pi * const.alpha_fsc) / (wavelength * 1e-9)**2) * (1 - 1 / (beta * n1)**2)

    return dNdLdw



# BSD 3-Clause License
#
# Copyright (c) 2021, Austin Cummings
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# TODO: calculate X offset based on energy of shower in sim file

''' Spatial Cherenkov Profile

.. autosummary::
   :toctree:
   :recursive:

   cherenkov_from_particle_profile

'''
import sys
import numpy as np
import matplotlib.pyplot as plt
from scipy import interpolate

from . import io_func
from . import constants as const
from . import atmosphere as atm
from . import geometry as geo
from . import cher_func as chf
from . import electron_func as elef
from . import fit_func as fitf
from . import plot

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

np.seterr(divide='ignore', invalid='ignore')
np.set_printoptions(threshold=sys.maxsize)


def cherenkov_from_particle_profile(detector_altitude,
                                    energy, theta_D,
                                    L_particle, L_detector, L_atmosphere,
                                    atmo, mag_field, run_set,
                                    plots
                                    ):
    """Calculates Spatial Cherenkov Profile given geometric constraints

    Parameters
    ----------
    detector_altitude : float
        Altitude of the detection plane in km
    energy : float
        energy of the shower in eV
    theta_D : float
        shower angle in degree defined as detector viewing angle
    L_particle : float
        propagated distance of the particle (in km) before interaction
    L_detector : float
        path lengths to the detector in km
    L_atmosphere : float
        path lengths the top of the atmosphere
    atmo : configparse_section Obj
        contains atmospheric settings: scattering, alpha_value file
    mag_field : configparse_section Obj
        conatins magnetic field settings: use, strength and angle
    run_set : configparse_section Obj
        contains all run specific settings like filenames, binnings and boundaries
    plots : str
        Description of which type of plots to make: none, general or all

    Returns
    -------
    dict
        Dictionary of simulation results
    """
    # Calculating slant depth profile as a function of propagation distance
    slant_depth_data = atm.slant_X_from_L_table([L_particle, min([L_detector, L_atmosphere])], detector_altitude, theta_D)

    # Creating interpolation objects to get slant depth from propagation distance and vice-versa
    L_from_slant_depth = interpolate.interp1d(slant_depth_data[1], slant_depth_data[0], fill_value='extrapolate')

    ################################################################
    # Initializing Atmosphere

    # Initializing the slant depth bins to calculate photon content
    max_x = float(run_set.getlist('slantDepth')[1])
    x_bins = int(run_set.getlist('slantDepth')[0])
    if max_x > 0:
        slant_depths = np.linspace(0, min(slant_depth_data[1][-1], max_x), x_bins)
    else:
        slant_depths = np.linspace(0, slant_depth_data[1][-1], x_bins)

    dx = slant_depths[1] - slant_depths[0]

    # Deleting the slant depth data object to save memory
    del slant_depth_data

    # Calculating the distance along the trajectory for each slant depth value
    Ls = L_from_slant_depth(slant_depths)

    # Calculating the altitude of each slant depth value
    zs = geo.z_from_L(Ls, theta_D, detector_altitude)

    # Integral calculation for time it takes photon to reach detector plane along shower axis
    # Take points along the path from the emission point to the detection plane
    Ls_t_value_calc = np.array([np.linspace(L_bottom, L_detector, 1000) for L_bottom in Ls])

    # Calculate the index of refraction at these points (using a reference 450nm wavelength)
    zs_t_value_calc = geo.z_from_L(Ls_t_value_calc, theta_D, detector_altitude)
    ns_t_value_calc = np.array([atm.index_of_refraction(zs_t_value_calc[i], 450) for i in range(0, len(Ls))])

    # Integrate the traversal time taking into account the changing speed of light and clear unneeded values
    t0s = np.trapz(ns_t_value_calc / const.cl, Ls_t_value_calc, axis=1)
    del Ls_t_value_calc, zs_t_value_calc, ns_t_value_calc

    # Initialize Wavelength Grid
    wavelength_bins, wavelength_lower, wavelength_upper = run_set.getlist('wavelength')
    wavelengths = np.linspace(float(wavelength_lower), float(wavelength_upper), int(wavelength_bins))

    # Loading in extinction coefficient from external file
    if atmo.get('alphaFile') != '':
        extinction_coefficient_file = atmo.get('alphaFile')
    else:
        extinction_coefficient_file = files("EASCherSim.data") / \
            "alpha_values_typical_atmosphere_updated.txt"
    alpha = atm.atmospheric_extinction_coefficient(extinction_coefficient_file)

    # Calculate the transmission coefficient for each emission point
    if atmo.getboolean('scattering'):
        if Ls[-1] < L_detector:
            extra_scattering = atm.optical_depth(alpha[0], alpha[1], alpha[2],
                                                 wavelengths,
                                                 np.linspace(Ls[-1], L_detector, 1000), detector_altitude,
                                                 theta_D,
                                                 False)[:, 0]

        else:
            extra_scattering = np.zeros(len(wavelengths))

        photon_transmission = np.exp(-atm.optical_depth(alpha[0], alpha[1], alpha[2],
                                     wavelengths,
                                     Ls, detector_altitude,
                                     theta_D,
                                     False) - extra_scattering[:, None])

    else:
        photon_transmission = np.ones((len(wavelengths), len(Ls)))

    del alpha

    ################################################################
    # Initializing shower

    # Calculating the number of electrons and positrons for each point in the shower
    if run_set.getboolean('useGreisenParametrization'):
        charged_parts = elef.Greisen_Longitudinal_Profile(energy, slant_depths)
    else:
        if run_set.get('protoShowerPofile') == '':
            filename = str(files("EASCherSim.data") / "Example_CORSIKA_proton_1e17eV.long")
            simulated_energy = 1e17
        else:
            filename = run_set.get('protoShowerPofile')
            simulated_energy = run_set.getfloat('protoShowerEnergy')
        charged_parts = elef.CORSIKA_interpolation(energy, slant_depths, filename, simulated_energy)
    # Calculating the position of the shower maximum, and the corresponding shower ages and radiation lenghts
    shower_xmax = slant_depths[np.argmax(charged_parts)]
    shower_age = 3 / (1 + 2 * shower_xmax / slant_depths)
    radiation_lengths = (slant_depths - shower_xmax) / 36.7

    ################################################################
    # Initializing Cherenkov photons

    # Calculating the threshold energy for Cherenkov production for each point,
    # using a reference 450nm wavelength
    Cherenkov_thresholds = chf.Cherenkov_threshold(zs, 450)

    # Calculating the bins in electron energy within which to generate Cherenkov emission
    # Spaced from the local Cherenkov threshold at each point to 0.1TeV (captures nearly
    # all of the charged particle behavior for all shower ages) in 30 bins
    electron_energy_bins = int(run_set.getlist('elEnergy')[0])
    upper_bound_electron_energy = float(run_set.getlist('elEnergy')[1])
    electron_energy_spacing = np.array([np.logspace(np.log10(thresh),
                                        np.log10(upper_bound_electron_energy),
                                        electron_energy_bins)
                                        for thresh in Cherenkov_thresholds])

    # Calculating the midpoint of the electorn energy bins and the fraction of electrons inside each bin
    electron_energies = electron_energy_spacing[:, :-1] + np.diff(electron_energy_spacing, axis=1) / 2
    electron_fractions = -np.diff(elef.Hillas_Energy_Distribution(electron_energy_spacing, shower_age), axis=1)

    # Calculate the total number of electrons contributing to Cherenkov emission per bin in energy and slant depth
    charged_parts_above_threshold = charged_parts[:, None] * electron_fractions

    # Calculating the total Cherenkov photons produced by electrons above the Cherenkov threshold
    total_Cherenkov_photons = (1 / atm.atmos_density(zs)[:, None, None]) * \
        1e-2 * \
        chf.Cherenkov_photons(zs, wavelengths, elef.particle_velocity(electron_energies)) * \
        charged_parts_above_threshold[:, :, None]

    # Applying atmospheric extinction to the propagated Cherenkov photons
    transmitted_Cherenkov_photons = total_Cherenkov_photons * photon_transmission.T[:, None, :]

    # Sum the Cherenkov photons in energy
    E_integrated_photons = np.sum(transmitted_Cherenkov_photons, axis=1)

    # Integrate the summed Cherenkov photons across slant depth
    X_integrated_photons = np.trapz(E_integrated_photons, slant_depths, axis=0)

    # Integrate the Cherenkov photons across wavelength
    W_integrated_photons = dx * np.trapz(transmitted_Cherenkov_photons, wavelengths * 1e-9, axis=2)

    ################################################################
    # Initialization for the angular distributions of electrons

    # Initializing the scaled angular bins to use for the electron angular distribution
    bins, start, stop = run_set.getlist('elZenith')
    scaled_angle_u = np.logspace(float(start), float(stop), num=int(bins))

    # Initializing the bins in azimuth about which to calculate photon content
    azimuth_bins = run_set.getint('ringAzimuthNumBin')
    Cherenkov_ring_azimuth = np.linspace(0, 2 * np.pi, azimuth_bins)

    ################################################################
    # Initializing the arrays to be used for the spatial, timing, and angular distributions of the arriving photons

    spatial_counts, time_counts, angle_counts = 0, 0 , 0

    spatial_bins = int(run_set.getlist('phSpatial')[0])
    bins, start, stop = run_set.getlist('phTime')
    time_bins = np.logspace(float(start), float(stop), num=int(bins))
    bins, start, stop = run_set.getlist('phZenith')
    angle_bins = np.logspace(float(start), float(stop), num=int(bins))

    # If generating plots, initialize the 2D spatial array
    if plot.Plot[plots] > 0:
        det_array = 0

    ################################################################
    # Beginning main loop
    # Loop over electron energy bins (start with the bin at the Cherenkov threshold for each altitude and move up to considered upper bound)

    with io_func.get_progress() as progress:
        io_func.flog(f'''Computing the Cherenkov Photon properties at an altitude of '''
                     f'''{detector_altitude}km for one EAS \n''')
        task = progress.add_task(f"1.Shower({energy}eV,{theta_D}deg,{L_particle}km)",
                                 total=electron_energies.shape[1])
        for i in range(0, electron_energies.shape[1]):
            # Selecting the electron energies and the Cherenkov photons emitted at this energy
            electron_energy, transmitted_photons = electron_energies[:, i], W_integrated_photons[:, i]

            # Calculate Lorentz factor and particle relative velocity to the speed of light
            electron_velocity = elef.particle_velocity(electron_energy)

            # Calculate the Cherenkov angle for each point and electron energy for a reference wavelength of 450nm
            Cherenkov_angles = chf.Cherenkov_angle(zs, 450, electron_velocity)

            # Setting NaN values equal to 0
            undefined_Cherenkov_angle = np.isnan(Cherenkov_angles)
            Cherenkov_angles[undefined_Cherenkov_angle] = 0
            transmitted_photons[undefined_Cherenkov_angle] = 0

            # Calculate Zenith angles and dn values at angles
            [electron_zenith_angle, electron_angular_content] = elef.Hillas_Angular_Distribution(electron_energy,
                                                                                                 shower_age,
                                                                                                 scaled_angle_u)

            # Dividing the generated Cherenkov photons into the corresponding bins in zenith angle
            angular_transmitted_photons = transmitted_photons[None, :] * electron_angular_content

            # Calculating the coordinates of the projected Cherenkov ring on the detection plane
            [ellipse_x, ellipse_y] = geo.ellipse_coordinates(Ls, L_detector,
                                                             electron_zenith_angle,
                                                             Cherenkov_angles, Cherenkov_ring_azimuth)

            # Calculating the arrival angle of the photons on the detection plane
            photon_arrival_angles = (180 / np.pi) * \
                np.arctan(np.sqrt(ellipse_x**2 + ellipse_y**2) / (L_detector - Ls[None, :, None]))

            # Randomly sampling lateral offsets and lag times for each Cherenkov photon generated
            lateral_offsets = elef.Lafebre_Lateral_Distribution(electron_energy, radiation_lengths, zs, ellipse_x)
            lag_times = elef.Lafebre_Lag_Time_Distribution(electron_energy, radiation_lengths, zs, ellipse_x)

            # Smearing sampled lateral offsets by randomly sampled azimuth angle
            sampled_angles = np.random.uniform(0, 2 * np.pi, size=ellipse_x.shape)
            x_offsets, y_offsets = lateral_offsets * np.cos(sampled_angles), lateral_offsets * np.sin(sampled_angles)

            del sampled_angles, lateral_offsets

            # Adding the randomly sampled offsets to each point about the Cherenkov ring
            ellipse_x += x_offsets
            ellipse_y += y_offsets

            del x_offsets, y_offsets

            # Calculating the distance from the shower axis for each point (in order to consider azimuthal symmetry)
            r_coordinates = np.sqrt(ellipse_x**2 + ellipse_y**2)

            del ellipse_x, ellipse_y
            total_propagation_times = (Ls[None, :, None] / const.cl + t0s[None, :, None] / np.cos((np.pi / 180) * photon_arrival_angles) + lag_times) / 1e-9
            total_propagation_times -= np.nanmin(total_propagation_times)

            # Calculating the number of Cherenkov photons arriving on the detection plane at each point along the Cherenkov ring
            photons_on_plane = (1 / azimuth_bins) * angular_transmitted_photons[:, :, None] * np.ones(azimuth_bins)[None, None, :]

            # Ignoring points which do not intersect the plane
            nan_indexes = np.isnan(r_coordinates)
            r_coordinates[nan_indexes] = 0
            total_propagation_times[nan_indexes] = 0
            photon_arrival_angles[nan_indexes] = 0
            photons_on_plane[nan_indexes] = 0

            # The first iteration of the loop has the maximum photon content, so at this step, the bounds are determined for the 2D histograms
            # that will be carried out for futher iterations in the loop
            if i == 0:
                # Find r bins such that a certain percentage of the total photon content is captured
                # We recommend 30% for near ring behavior
                pct_cap = float(run_set.getlist('phSpatial')[1])
                r_percentiles = fitf.weighted_percentile(r_coordinates.flatten(), pct_cap, weights=photons_on_plane.flatten())
                r_bins = np.linspace(0, r_percentiles, spatial_bins)

                # Calculate the size of one spatial bin in m^2
                pixel_size = ((r_bins[1] - r_bins[0]) * 1000)**2

                # Initializing the arrays to store the upper and lower bounds for the timing and angular photon distributions
                time_bounds, zenith_bounds = [], []

                # Loop over the spatial bins
                for j in range(0, len(r_bins) - 1):
                    # Find coordinates which are inside each spatial bin
                    r_lower, r_upper = r_bins[j], r_bins[j + 1]
                    in_range = (r_coordinates >= r_lower) & (r_coordinates <= r_upper)
                    in_range_time, in_range_zenith, in_range_photons = total_propagation_times[in_range], photon_arrival_angles[in_range], photons_on_plane[in_range]

                    # Calculate the 1% and 99% weighted percentiles of the arrival time and arrival angle inside the bin
                    time_percentiles = fitf.weighted_percentile(in_range_time.flatten(), np.array([1, 99]), weights=in_range_photons.flatten())
                    zenith_percentiles = fitf.weighted_percentile(in_range_zenith.flatten(), np.array([1, 99]), weights=in_range_photons.flatten())

                    time_bounds.append(time_percentiles)
                    zenith_bounds.append(zenith_percentiles)

                # Calculate the lower bound time and zenith offsets as a function of r (in km)
                time_offset = interpolate.interp1d(r_bins[:-1], np.array(time_bounds)[:, 0], kind='cubic', bounds_error=False, fill_value=0)
                angle_offset = interpolate.interp1d(r_bins[:-1], np.array(zenith_bounds)[:, 0], kind='cubic', bounds_error=False, fill_value=0)

            # Plot the photon intensity (in photons/m^2) in 2 dimensions
            if plot.Plot[plots]:
                # Smear the calculated distances from shower axis about the axis random;y
                random_azimuth_angles = np.random.uniform(0, 2 * np.pi, size=r_coordinates.shape)
                x_coordinates, y_coordinates = r_coordinates * np.cos(random_azimuth_angles), r_coordinates * np.sin(random_azimuth_angles)

                # Shift the x coordinates by the calculated offset due to geomagnetic deflection (axis is arbitrary)
                if mag_field.getboolean('useField'):
                    x_coordinates += elef.geomagnetic_deflection_correction(Ls, L_detector,
                                                                            electron_energy,
                                                                            electron_zenith_angle,
                                                                            Cherenkov_ring_azimuth,
                                                                            mag_field.getfloat('strength'),
                                                                            mag_field.getfloat('angle'))

                # Update the 2D array stored in memory
                current_array = np.histogram2d(np.abs(x_coordinates.flatten()), np.abs(y_coordinates.flatten()), bins=(r_bins, r_bins), weights=photons_on_plane.flatten() / (4 * pixel_size))[0]
                det_array += current_array

                del random_azimuth_angles, x_coordinates, y_coordinates

            # Shifting the calculated values by the lower offset
            total_propagation_times -= time_offset(r_coordinates.flatten()).reshape(total_propagation_times.shape)
            photon_arrival_angles -= angle_offset(r_coordinates.flatten()).reshape(photon_arrival_angles.shape)

            # Calculating the 1D histogram for the spatial counts and the 2D histogram for the arrival times and arrival angles
            spatial_counts += np.histogram(r_coordinates.flatten(), r_bins, weights=photons_on_plane.flatten())[0]
            time_counts += np.histogram2d(r_coordinates.flatten(), total_propagation_times.flatten(), bins=(r_bins, time_bins), weights=photons_on_plane.flatten())[0]
            angle_counts += np.histogram2d(r_coordinates.flatten(), photon_arrival_angles.flatten(), bins=(r_bins, angle_bins), weights=photons_on_plane.flatten())[0]

            del r_coordinates, total_propagation_times, photon_arrival_angles, photons_on_plane
            progress.advance(task)
    io_func.flog("[bold blue]Shower 1 done! :heavy_check_mark:")
    # Plotting Wavelength, Angular, Time, and Spatial Distributions of the arriving Cherenkov photons
    if plot.Plot[plots]:
        plot.plot_wavelength_distribution(wavelengths, X_integrated_photons)
        plot.plot_photon_angular_spread(r_bins, angle_bins, angle_counts)
        plot.plot_photon_time_spread(r_bins, time_bins, time_counts)
        plot.plot_3D_spatial_distribution(r_bins, det_array)

    # Correcting the spatial counts to account for the area of the ring (photons/m^2)
    ring_area = np.pi * ((r_bins[1:] * 1000)**2 - (r_bins[:-1] * 1000)**2)
    spatial_counts = spatial_counts / (ring_area)

    ################################################################

    cher_res = {
        "wavelengths" : wavelengths,
        "wavelength_counts" : X_integrated_photons[:-1],
        "r_bins" : r_bins,
        "r_counts" : spatial_counts,
        "time_bins" : time_bins,
        "time_counts" : time_counts,
        "lower_times" : np.array(time_bounds)[:, 0],
        "angle_bins" : angle_bins,
        "angle_counts" : angle_counts,
        "lower_angles" : np.array(zenith_bounds)[:, 0],
    }

    if plot.Plot[plots]:
        plt.show()

    return cher_res
